#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdint.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <pwd.h>
#include <grp.h>
#include <err.h>

#include "common.h"
#include "macros.h"
#include "misc.h"
#include "xmalloc.h"
#include "failures.h"
#include "buffer.h"

#define DEV_NULL	"/dev/null"

void message(int type, const char *filename, int line, const char *fmt, ...)
{
	va_list ap;
	int n;
	char fmtbuf[256];
	char buffer[1024];
	char *text;

	va_start(ap, fmt);

	switch (type) {
	case 0:
		text = "warning";
		break;
	case 1:
		text = "error";
		break;
	case 2:
		text = "note";
		break;
	default:
		text = "???";
		break;
	}

	n = 0;

	if (filename != NULL)
		n = snprintf(fmtbuf, sizeof(fmtbuf), "%s:%d: ", filename, line);

	snprintf(fmtbuf+n, sizeof(fmtbuf)-n, "%s: %s" CRLF, text, fmt);

	n = vsnprintf(buffer, sizeof(buffer), fmtbuf, ap);

	if (n > 0) {
		write(STDERR_FILENO, buffer, n);
	} else {
		SHOULDNT_REACH();
	}

	va_end(ap);
}

ssize_t write_full(int fd, const char *buf, size_t size)
{
	ssize_t ret, left;
	const char *ptr;

	ret = 0;
	ptr = buf;
	left = size;

	while (left) {
		ret = write(fd, ptr, left);
		if (ret < 0) {
			if (errno == EINTR)
				continue;
			return -1;
		} else {
			left -= ret;
			ptr += ret;
		}
	}

	return size;
}

ret_t open_file_ro(const char *file_name, int *fd)
{
	struct stat st;
	int fdesc;

	return_val_if_fail(file_name != NULL, ret_err);

	if (access(file_name, R_OK) != 0)
		return ret_no_access;

	if (stat(file_name,&st) != 0)
		return ret_err;

	if (!S_ISREG(st.st_mode) || st.st_size < 0)
		return ret_not_file;

	fdesc = open(file_name, O_RDONLY);
	if (fdesc == -1)
		return ret_err;

	*fd = fdesc;
	return ret_ok;
}

ret_t pipe_open(const char *file_name, int mode, int *fdres)
{
	struct stat st;
	int res, fd, flags;

	return_val_if_fail(file_name != NULL, ret_err);
	return_val_if_fail(fdres != NULL, ret_err);

	fd = open(file_name, mode);
	if (fd == -1) {
		fprintf(stderr, "open failed: `%s' %s\n", file_name,
							strerror(errno));
		return ret_err;
	}
		
	res = fstat(fd, &st);
	if (res == -1) {
		fprintf(stderr, "stat failed: `%s' %s\n", file_name,
							strerror(errno));
		goto err;
	}

	if (!S_ISFIFO(st.st_mode)) {
		fprintf(stderr, "input file is not a pipe `%s'\n", file_name);
		goto err;
	}

	flags = fcntl(fd, F_GETFL);
	res = fcntl(fd, F_SETFL, flags | O_NONBLOCK);
	if (res == -1) {
		fprintf(stderr, "can't set nonblocking mode `%s'\n", file_name);
		goto err;
	}

	*fdres = fd;

	return ret_ok;

err:
	close(fd);	
	*fdres = -1;
	return ret_err;
}

ret_t close_file(int *fd)
{
	return_val_if_fail(*fd > 0, ret_err);

	if (close(*fd) != 0)
		return ret_err;

	*fd = -1;
	return ret_ok;
}

ret_t file_length(int fd, size_t *length)
{
	struct stat st;
	off_t len;
	int ret;

	return_val_if_fail(fd > 0, ret_err);

	ret = fstat(fd, &st);
	if (ret == -1)
		return ret_err;

	len = st.st_size;
	if (len == 0)
		return ret_err;
	/* not sure that off_t and size_t the same bits length */
	*length = len;

	return ret_ok;
}

ret_t mmap_file(const int fd, void **mmap_addr, size_t length)
{

	void *ptr;

	return_val_if_fail(fd > 0, ret_err);

	ptr = mmap(NULL, length, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
	if (ptr == (void *)-1)
		return ret_err;

	*mmap_addr = ptr;
	return ret_ok;
}

ret_t munmap_file(void **addr, const size_t length)
{
	return_val_if_fail(*addr != NULL, ret_err);

	if (munmap(*addr, length) != 0)
		return ret_err;

	*addr = NULL;
	return ret_ok;
}

char *extract_string(struct buffer *buf, char *str, size_t max_len, size_t *_len, int delim)
{
	char *beg, *end;
	size_t len;

	beg = buffer_ptr(buf);

	end = strchr(beg, delim);
	if (end == NULL)
		return NULL;

	*end = '\0';
	len = end-beg;

	if (len >= max_len)
		return NULL;	

	strncpy(str, beg, len+1);
	*end = delim;
	
	if (_len != NULL)
		*_len = len;

	return str;
}

int str_to_uint(const char *str, char **endptr, unsigned int *num, int base)
{
	unsigned long n;
	char *end;

	errno = 0;
	n = strtoul(str, &end, base);

	if ((errno == ERANGE && (n == 0 || n == ULONG_MAX)) || 
			errno == EINVAL || errno != 0)
		return -1;

	*num = n;

	if (endptr != NULL && *endptr != NULL)
		*endptr = end;

	return 0;
}

int xsnprintf(char *str, size_t size, const char *fmt, ...)
{
	va_list ap;
	int res;
	
	va_start(ap, fmt);

	res = vsnprintf(str, size, fmt, ap);
	if (res < 0)
		errx(1, "vsnprintf res=%d", res);

	va_end(ap);

	return res;
}

int mem_touch_and_lock(const void *addr, size_t len)
{
	int res, page_size, i;
	unsigned char *buf;

	return_val_if_fail(addr != NULL, -1);
	return_val_if_fail(len > 0, -1);

	buf = (unsigned char *)addr;

	page_size = getpagesize();

	for (i = 0; i < len; i += page_size) {
		res = buf[i];
		buf[i] = res;
	}

	res = mlock(addr, len);
	if (res == -1)
		return -1;

	return 0;
}

int ascii_to_hex(char c)
{
	int res;

	if (isupper(c))
		c = tolower(c);

	if (c >= '0' && c <= '9')
		res = c - '0';
	else if (c >= 'a' && c <= 'f')
		res = (c - 'a') + 10;
	else
		res = 0;

	return res;
}

char hex_to_ascii(int value)
{
	char c;

	if (value >= 0 && value <= 9)
		c = '0' + value;
	else if (value >= 0xa && value <= 0xf)
		c = 'a' + (value - 10);
	else
		c = '0';

	return c;
}

int str_isnumber(const char *str)
{
	if (!str || *str == 0)
		return 0;

	while (*str && isdigit(*str))
		str++;

	return (*str == 0);
}

void sanitize_stdio()
{
	int res;
	int i, devnull;

	devnull = open(DEV_NULL, O_RDWR);

	if (devnull == -1)
		xerror(1, "open() '%s' failed: %s", DEV_NULL, strerror(errno));

	i = devnull;

	while (++i <= 2) {
		res = fcntl(i, F_GETFL);
		if (res >= 0)
			continue;
		res = dup2(devnull, i);
		if (res == -1)
			xerror(1, "dup2() can't dup fd=%d: %s", i, strerror(errno));
	}

	if (devnull > 2)
		close(devnull);
}

void daemonize()
{
	int i, devnull;
	int res;
	pid_t pid;

	chdir("/");
	umask(0220);
	pid = fork();
	if (pid > 0)
		exit(0);
	if (pid < 0)
		xerror(1, "fork() failed: %s", strerror(errno));
	setsid();
	devnull = open(DEV_NULL, O_RDWR | O_NOCTTY);
	if (devnull == -1)
		xerror(1, "can't open() '%s': %s", DEV_NULL, strerror(errno));
	i = 0;
	do {
		res = close(i);
		if (res == -1)
			xwarning("close() failed: %s", strerror(errno));
		res = dup2(devnull, i);
		if (res == -1)
			xwarning("dup2() failed: %s", strerror(errno));
	} while (++i <= 2);
	close(devnull);
}

int getpw_by_name_or_uid(const char *str, struct passwd *pwd)
{
	struct passwd *ptr = NULL;

	if (!str)
		return -1;

	if (str_isnumber(str)) {
		do {
			errno = 0;
			ptr = getpwuid(atoi(str));
		} while (ptr == NULL && errno == EINTR);
	} else {
		do {
			errno = 0;
			ptr = getpwnam(str);
		} while (ptr == NULL && errno == EINTR);
	}
	if (ptr) {
		if (pwd != NULL)
			*pwd = *ptr;
		return 0;
	}
	return -1;
}

int getgr_by_name_or_gid(const char *str, struct group *grp)
{
	struct group *ptr = NULL;

	if (!str)
		return -1;

	if (str_isnumber(str)) {
		do {
			errno = 0;
			ptr = getgrgid(atoi(str));
		} while (ptr == NULL && errno == EINTR);
	} else {
		do {
			errno = 0;
			ptr = getgrnam(str);
		} while (ptr == NULL && errno == EINTR);
	}
	if (ptr) {
		if (grp != NULL)
			*grp = *ptr;
		return 0;
	}
	return -1;
}

/* helper to obtain canonical absolute path */
char *canonize_path(const char *str)
{
	size_t len;
	char *p, *q, *cwd, *res;
	char *path;

	if (!str)
		return NULL;

	if (*str == '/') {
		len = strlen(str);
		res = xmalloc0(len + 1);
		path = xstrdup(str);
		res[0] = '/';
		p = res;
	} else {
		cwd = xgetcwd();
		len = strlen(cwd) + strlen(str) + 2;
		res = xmalloc0(len);
		path = xmalloc0(len);
		snprintf(path, len, "%s/%s", cwd, str);
		p = res;
		xfree(cwd);
	}

	q = path;

	while (*q) {
		if (q[0] == '/') {
			if (q[1] && q[1] == '/') {
				q += 1;
				continue;
			}
			if (q[1] && q[1] == '.') {
				if (q[2] == 0 || q[2] == '/') {
					q += 2;
					continue;
				}
				if (q[2] && q[2] == '.') {
					if (q[3] == 0 || q[3] == '/') {
						if (p != res && *p == '/')
							*p-- = 0;
						while (p != res && *p != '/')
							*p-- = 0;
						q += 3;
						continue;
					}
				}
			}
		}
		*p++ = *q++;
	}

	xfree(path);

	return res;
}
